import games
import random
import util

class GameAgent:
    """ Abstract GameAgent class. Your agents should inherit from this class 
        and override the getAction method.  """
  
    def getAction(self, gameState):
        """  Return a move in gameState.actions(), where gameState is an instance of a subclass of GameState. """
        abstract


class RandomGameAgent(GameAgent):
    """  A game agent that chooses a legal action at random."""   
   
    def getAction(self, currentState):
        """  Choose a legal action at random."""   
        return random.choice(currentState.actions())


class HumanGameAgent(GameAgent):
    """A game agent that asks the user what action to take. """ 

    def getAction(self, currentState):
        """  Print useful information, and then ask the user what action to take. """   
        print "\n"
        if currentState.getPreviousMove():
            print "\nOpponent's move was %s" % str(currentState.getPreviousMove())
        print currentState
    
        legalActions = currentState.actions()
        try: action = input("\nYour move? ")
        except (KeyboardInterrupt, SystemExit): raise
        except: action = None
        while action not in legalActions:
            print "That is not a legal move.  Please enter one of %s" % legalActions
            try: action = input("\nYour move? ")
            except (KeyboardInterrupt, SystemExit): raise
            except: action = None
        return action



class SimpleTTTGameAgent(GameAgent):
    """  
         A simple agent that plays Tic Tac Toe.  By default, it tries to go in the center
         and otherwise plays in the first available squre.  
         
         You should modify it to do something more intelligent.
    """   
   
    def getAction(self, currentState):
        """  
            currentState is an instance of TicTacToeGameState.  
            By default, this agent tries to move in the center square, or picks the first
            open square if the center is already taken.
            
            TODO (Q1): Modify it to do something more intelligent, ideally 
             (but not necessarily) so that it never loses. 
        """   
        # TODO(Q1): improve this agent
        board = currentState.board
        if board[1][1] == " ":
            return (1,1)
        
        for y in range(3):
            for x in range(3):
                if board[y][x] == " ":
                    return (x,y)





class UtilityDirectedGameAgent(GameAgent):
    """ A game agent that chooses the action with the highest utility  """ 
    def __init__(self, searcher):
        self.searcher = searcher;

    def getAction(self, currentState):
        bestAction, value = self.searcher.getBestActionAndValue(currentState)
        print "Got state value %s with best action %s" % (value, bestAction)
        return bestAction  


class GameTreeSearcher:    
    """ Interface for classes that search the game tree exhaustively, 
        returning a maximum-utility action and its value (or a bound thereupon) """

    def getBestActionAndValue(self, currentState):
        """ Return the best action and its value in currentState for currentState.currentPlayer() """
        abstract
    



    
class Minimax(GameTreeSearcher):
    """
        A game agent that picks the best action by exhaustively searching the game tree,
        and backing up values according to the minimax algorithm.
        
        TODO (Q2): Fill in the implementation of this class.
        When complete, it should be able to prove that Tic Tac Toe is a draw 
        from the starting position.
        
        You can either implement the classical version of minimax that uses two functions 
        (a minimizing function for min-player and a maximizing function for max-player),
        or take advantage of the fact that we are only interested in Zero-sum games
        to implement "Negamax", which is simpler and requires only one recursive function.
    """ 

    def getBestActionAndValue(self, currentState):
        """ Return the best action and its value in currentState for currentState.currentPlayer() """
        # TODO (Q2): Complete this method.
        return None, 0 

    
class GraphMinimax(Minimax):
    """
        A game agent that picks the best action by exhaustively searching the game graph,
        and backing up values according to the minimax algorithm.
        
        TODO (Q3): Fill in the implementation of this class.
        When complete, it should be able to prove that Tic Tac Toe is a draw 
        from the starting position, significantly faster than ordinary Minimax.
        
        You may find it useful to call or otherwise uilitize your ordinary Minimax 
        version of getBestActionAndValue as a subroutine.  You can access this method
        as Minimax.getBestActionAndValue
    """ 
    
    def getBestActionAndValue(self, currentState):
        """ Return the best action and its value in currentState for currentState.currentPlayer() """
        self.transpositionTable = {}
        # TODO (Q3): Complete this method.
        return None, 0 




class AlphaBeta(GameTreeSearcher):
    """
        A game agent that picks the best action by exhaustively searching the game tree,
        backing up values according to the minimax algorithm, and pruning unnecessary 
        branches using the Alpha-Beta Pruning algorithm
        
        TODO (Q4): Fill in the implementation of this class.
        When complete, it should be able to prove that Tic Tac Toe is a draw 
        from the starting position, significantly faster than ordinary Minimax.
        
        You can assume that all utilities are in the range [-1, 1].
        If you do, please don't forget this when you implement your heuristic functions later.
        
        As for Minimax, you can opt for a mutually recursive two-function approach, 
        or take advantage of the fact that we are only interested in zero-sum games
        to implement a single recursive "Negamax" version. 
    """ 

    def getBestActionAndValue(self, currentState):
        """ Return the best action and its value in currentState for currentState.currentPlayer() """
        # TODO (Q4): Write this method
        return None, 0






class FixedDepthHeuristicGameAgent(UtilityDirectedGameAgent):
    """A game agent that chooses the action with the max expected utility
       according to heiristicEvalFn, after looking ahead by searchDepth ply.  
       You will create instances of this class to use your heuristic functions,
       but you don't need to understand how it works in detail. """ 

    def __init__(self, searcher, heuristicEvalFn, searchDepth):
        self.searcher = searcher;
        self.heuristicEvalFn = heuristicEvalFn
        self.searchDepth = searchDepth

    def getAction(self, currentState):
        wrappedState = games.FixedDepthHeuristicGameStateAdapter(currentState, self.heuristicEvalFn, self.searchDepth)
        return UtilityDirectedGameAgent.getAction(self, wrappedState)




def uniformSquareValue(size, x, y):
    """ A function that gives all squares an equal weight of 1 """
    return 1


def edgeCornerSquareValue(size, x, y):
    """ 
        A function that returns different values for interior, edge, and corner squares on
        an othello board (0,size-1),  ... ,  (size-1, size-1)
                           ....................
                         (0,0),      (1,0), ... (size-1, 0)

        TODO (Q6): Complete this function so that it gives higher value to corner squares 
        than edge squares than interior squares, experimenting with different values to see
        which performs best.  Your function should work for any value of size.
    """
    # TODO (Q6): complete this function
    return 1



def pieceCountOthelloHeuristic(state, squareValueFn = uniformSquareValue):
    """ 
        A heuristic evaluation function for Othello that values a position
        by the difference in the number of pieces of the two players, weighted
        by position according to squareValueFn.  By default, uniformSquareValue
        is used, which assigns all squares a weight of 1.
    """
    size = state.boardSize
    currentPlayer = state.currentPlayer()
    board = state.board
    
    runningTotal = 0.0
    maxPoss = 0.0
    edges = (0, size-1)
    for y in range(size):
        for x in range(size):
            square = board[y][x]
            value = squareValueFn(size,x,y)
            maxPoss += value
            if square != " ":                
                if square == currentPlayer: runningTotal += value
                else:                       runningTotal -= value
            
    return runningTotal / maxPoss 


def edgeCornerOthelloHeuristic(state): 
    """ An Othello heuristic that counts pieces weighted by edgeCornerSquareValue(size, x, y) """
    return pieceCountOthelloHeuristic(state, edgeCornerSquareValue)




def perfectNimHeuristic(state):
    """ 
        A perfect heuristic for Nim, which returns the exact value of this state
        for state.currentPlayer() in constant time.
        
        TODO (Q7): Complete this function.
    """
    # TODO (Q7): Complete this function.
    return 0

    
